import datetime
import time
from typing import Type

from auto_easy.models.geometry import *


class ItemT(Enum):
    Default = "默认"
    Role = "角色"
    GuidePost = "路标"
    Monster = "怪物"
    Goods = "物品"


class YoloItem:
    def __init__(self, name, score, match_box, src_box):
        self.params_list = [name, score, match_box, src_box]
        self.name = name
        self.score = score
        self.box = match_box
        self.src_box = src_box
        self.mock = False
        self.mock_mid_point = Point(0, 0)

    # 默认返回识别区域中心点，派生类可以覆盖
    @property
    def middle_point(self) -> Point:
        if self.mock:
            return self.mock_mid_point
        return self.box.get_mid_point()

    @property
    def itype(self):
        return ItemT.Default

    def __str__(self):
        return "name({})-type({})-class({})-mid_point({})".format(self.name, self.itype.value,
                                                                  self.__class__.__name__,
                                                                  str(self.middle_point))

    def cal_farthest(self, items):
        target = items[0]
        max_dis = self.middle_point.get_distance(items[0].middle_point)
        for item in items:
            dis = self.middle_point.get_distance(item.middle_point)
            if max_dis < dis:
                target = item
                max_dis = dis
        return target

    def cal_closest(self, items):
        target = items[0]
        min_dis = self.middle_point.get_distance(items[0].middle_point)
        for item in items:
            dis = self.middle_point.get_distance(item.middle_point)
            if min_dis > dis:
                target = item
                min_dis = dis
        return target

    def cal_left(self, items):
        res = []
        for item in items:
            if item.middle_point.x <= self.middle_point.x:
                res.append(item)
        return res

    def cal_right(self, items):
        res = []
        for item in items:
            if item.middle_point.x > self.middle_point.x:
                res.append(item)
        return res

    def move(self, x, y):
        self.box.move(x, y)

    def is_class(self, cls: type):
        if not isinstance(cls, type):
            raise Exception('AIItemBase.isinstance cls must class type')
        mro_list = self.__class__.__mro__
        mro_name_list = [c.__name__ for c in mro_list]
        return cls.__name__ in mro_name_list

    def in_classes(self, cls_list: List[type]):
        for cls in cls_list:
            if self.is_class(cls):
                return True
        return False

    @staticmethod
    def can_new_obj(input: str) -> bool:
        return True


class MYoloItem:
    def __init__(self, items: list[YoloItem], model_name=''):
        self.model_name = model_name
        self.items = items
        self.create_ts = time.time()

    @property
    def is_detected(self):
        return len(self.items) > 0

    def cal_item_cnt(self, name):
        return len(self.get_items_by_name(name))

    def contain_item(self, name):
        return len(self.get_items_by_name(name)) > 0

    def exists_type(self, itype: ItemT) -> bool:
        return len(self.get_items_by_type(itype)) > 0

    def exists_types(self, itypes: list[ItemT]) -> bool:
        flag = True
        for t in itypes:
            if len(self.get_items_by_type(t)) == 0:
                flag = False
        return flag

    def check(self, include=[], exclude=[]):
        cls_include = []
        type_include = []
        for v in include:
            if isinstance(v, ItemT):
                type_include.append(v)
            elif isinstance(v, type):
                cls_include.append(v)
            else:
                raise Exception(f'only check cls/type, include: {include}')
        if len(type_include) > 0 and not self.exists_types(type_include):
            return False
        if len(cls_include) > 0 and not self.exists_cls_list(cls_include):
            return False

        cls_exclude = []
        type_exclude = []
        for v in exclude:
            if isinstance(v, ItemT):
                type_exclude.append(v)
            elif isinstance(v, type):
                cls_exclude.append(v)
            else:
                raise Exception(f'only check cls/type, include: {exclude}')
        if len(cls_exclude) > 0 and self.exists_cls_list(cls_exclude, min_num=1):
            return False

        if len(type_exclude) > 0 and len(self.get_items_by_types(type_exclude)) > 0:
            return False
        return True

    def exists_cls(self, cls: type) -> bool:
        return self.get_item_by_cls(cls) is not None

    def exists_cls_list(self, cls_list: list[type], min_num=-1) -> bool:
        cnt = 0
        for cls in cls_list:
            if not self.exists_cls(cls):
                return False
            cnt += 1
            if min_num > 0 and cnt >= min_num:
                return True
        return True

    def get_item_by_cls(self, cls: type) -> YoloItem:
        ls = self.get_items_by_cls(cls)
        return ls[0] if len(ls) > 0 else None

    def get_item_by_cls_list(self, order_cls_list: list[type]) -> YoloItem:
        for cls in order_cls_list:
            item = self.get_item_by_cls(cls)
            if item is not None:
                return item
        return None

    def get_items_by_cls(self, cls: type) -> list[YoloItem]:
        ls = []
        for item in self.items:
            if item.is_class(cls):
                ls.append(item)
        return ls

    def get_items_by_types(self, itype_list: list[ItemT]) -> list[YoloItem]:
        ls = []
        for item in self.items:
            if item.itype in itype_list:
                ls.append(item)
        return ls

    def get_items_by_type(self, itype: ItemT) -> list[YoloItem]:
        ls = []
        for item in self.items:
            if item.itype == itype:
                ls.append(item)
        return ls

    def get_items_by_name(self, name) -> list[YoloItem]:
        if name == '':
            return self.items
        ls = []
        for item in self.items:
            if item.name == name:
                ls.append(item)
        return ls

    def get_item_by_type(self, itype: ItemT, cls_order_list: list[Type] = None) -> YoloItem:
        items = self.get_items_by_type(itype)
        if len(items) == 0:
            return None

        if cls_order_list is not None:
            for cls in cls_order_list:
                for item in items:
                    if isinstance(item, cls):
                        return item
        return items[0]

    def get_item_by_name(self, name) -> YoloItem:
        items = self.get_items_by_name(name)
        if len(items) == 0:
            return None
        return items[0]

    def get_item_types(self) -> list[ItemT]:
        ts = {}
        for item in self.items:
            ts[item.itype] = 0
        return ts.keys()

    def get_items_name(self) -> list[str]:
        set_val = set()
        for item in self.items:
            set_val.add(str(item.itype.value))
        return list(sorted(set_val))

    def __str__(self):
        ls = []
        for item in self.items:
            ls.append(str(item))
        ts_format = datetime.datetime.fromtimestamp(self.create_ts)
        return 'model[{}], time[{}]: '.format(self.model_name, ts_format) + ';'.join(ls)

    def print_simple_info(self):
        ls = []
        for item in self.items:
            ls.append('{}{}'.format(item.name, item.middle_point))
        return '|'.join(sorted(ls))

    def print_info(self):
        ls = []
        for item in self.items:
            ls.append('{}&{}&{}'.format(item.name, item.middle_point, item.box))
        return ';'.join(sorted(ls))


def cvt_items2points(items: list[YoloItem]) -> list[Point]:
    points = []
    for item in items:
        points.append(item.middle_point)
    return points


def cnt_point(points, f) -> int:
    cnt = 0
    for point in points:
        if f(point):
            cnt += 1
    return cnt


if __name__ == '__main__':
    pass
    # # cnt_point(points, lambda point: point.x > role.middle_point.x)
    # s = "哥布林,0.869,819,120,54,86|人物,0.741,521,293,15,20|哥布林,0.636,59,128,53,96"
    # s = "路标-移动,0.923,882,216,44,26|路标-三角,0.854,579,303,24,18|路标-三角,0.835,491,305,24,15|人物-叶子,0.822,180,158,20,21|人物-黑钻,0.801,156,161,18,17|建筑-小桥,0.785,254,266,425,74|路标-三角,0.771,764,305,21,14|路标-三角,0.751,308,305,21,15|路标-三角,0.750,398,305,24,15|路标-三角,0.684,947,305,14,14|物品-白色物品,0.631,791,222,54,18|路标-三角,0.607,672,306,20,12|人物-工会图标,0.606,54,162,15,12"
    # items = new_ai_base_item(s)
    #
    # res = AIResult(items, s)
    # print(res)
    #
    # print(check_distance([
    #     Point(494, 292),
    #     Point(494, 290),
    #     Point(385, 365),
    # ], user_max_x=20, user_max_y=20))

    # items = res.get_items_by_name('')
    # target = AIItemBase('t', 0.9, 820, 120, 55, 86)
    # closest = find_closest_point(items,target)
    # print( closest.name, closest.score)
    #
    # p1 = Point(400,600,'')
    # p2 = Point(800,1200,'')
    # print(gen_move_target_point(p1, p2).x)



import random


class Timeout:
    def __init__(self, seconds):
        """
        初始化函数，接受一个表示超时时间的参数seconds（单位：秒）
        """
        self.seconds = seconds
        self.start_time = time.time()
        self.check_times = 0

    def not_timeout(self):
        """
        对外函数，用于检查是否超过设定的超时时间，超过则返回False，未超过返回True
        """
        # 如果超时时间设置为0,允许调用一次
        self.check_times += 1
        if self.seconds == 0 and self.check_times == 1:
            return True
        current_time = time.time()
        if current_time - self.start_time > self.seconds:
            return False
        return True

    def is_timeout(self):
        return not self.not_timeout()


def calculate_average_milliseconds(time_array):
    """
    计算给定时间数组中时间对象的平均耗时，单位为毫秒。

    参数:
    time_array (list): 包含datetime.now()获取的时间对象的数组，假设时间对象成对出现表示开始和结束时间。

    返回:
    int: 平均耗时，单位为毫秒。
    """
    if len(time_array) == 0:
        return 0
    durations_in_milliseconds = []
    for i in range(len(time_array) - 1):
        start_time = time_array[i]
        end_time = time_array[i + 1]
        # 取出毫秒信息并计算时间差（单位：毫秒）
        time_diff = end_time - start_time
        duration_milliseconds = int(time_diff.total_seconds() * 1000)
        durations_in_milliseconds.append(duration_milliseconds)

    if durations_in_milliseconds:
        total_milliseconds = sum(durations_in_milliseconds)
        average_duration_milliseconds = total_milliseconds // len(durations_in_milliseconds)
        return average_duration_milliseconds
    return 0


def cost_ms(start) -> int:
    diff = time.time() - start
    return int(diff * 1000)


import functools
import time


def limit_to_one_true_per_x_seconds_decorator(x):
    """
    装饰器函数，使得被装饰的函数在主线程持续调用时，在最近x秒内只返回一次True。

    参数:
    x (float或int): 限定的时间间隔（秒）。
    """
    last_true_time = None

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal last_true_time
            current_time = time.time()
            if last_true_time is None or current_time - last_true_time > x:
                last_true_time = current_time
                return func(*args, **kwargs)
            return True

        return wrapper

    return decorator


@limit_to_one_true_per_x_seconds_decorator(1)
def one_sec_one_false(x):
    return False


def sleep_with_rand(sec, min_rate=0.8, max_rate=1.2):
    time.sleep(random.uniform(sec * min_rate, sec * max_rate))


def sleep_with_ms(ms, min_rate=1.0, max_rate=1.0):
    sec = ms / 1000
    time.sleep(random.uniform(sec * min_rate, sec * max_rate))


from datetime import datetime


def get_cur_timestr():
    now = datetime.now()
    time_str = now.strftime('%Y%m%d%H%M%S')
    return time_str


if __name__ == '__main__':
    print(get_cur_timestr())
    ############ 测试平均时间计算

    # 模拟存储时间字符串的列表，你可以根据实际情况替换这里的数据来源
    time_strings = [
        '无移动->有移动: 2024-12-11 15:52:05.597354',
        '无移动->有移动: 2024-12-11 15:52:07.195285',
        '无移动->有移动: 2024-12-11 15:52:08.965235',
        '无移动->有移动: 2024-12-11 15:52:10.421640',
        '无移动->有移动: 2024-12-11 15:52:12.023158',
        '无移动->有移动: 2024-12-11 15:52:13.639344',
        '无移动->有移动: 2024-12-11 15:52:15.234323',
        '无移动->有移动: 2024-12-11 15:52:16.844498',
        '无移动->有移动: 2024-12-11 15:52:18.443976',
        '无移动->有移动: 2024-12-11 15:52:20.045133',
        '无移动->有移动: 2024-12-11 15:52:21.660275',
        '无移动->有移动: 2024-12-11 15:52:23.249212',
        '无移动->有移动: 2024-12-11 15:52:24.869407',
        '无移动->有移动: 2024-12-11 15:52:26.463027',
        '无移动->有移动: 2024-12-11 15:52:28.069128',
        '无移动->有移动: 2024-12-11 15:52:29.664688',
        '无移动->有移动: 2024-12-11 15:52:31.270980',
        '无移动->有移动: 2024-12-11 15:52:32.865749',
    ]

    datetime_objects = []
    for time_string in time_strings:
        # 提取出时间部分的字符串
        time_part = time_string.split(": ", 1)[-1]
        # 将时间字符串转换为datetime对象
        dt = datetime.strptime(time_part, "%Y-%m-%d %H:%M:%S.%f")
        datetime_objects.append(dt)

    print(datetime_objects)

    # 模拟一个包含时间记录的数组，这里假设是成对出现，前一个是开始时间，后一个是结束时间，依次类推
    # time_records = [
    #     datetime(2024, 12, 11, 10, 0, 0, 10000),
    #     datetime(2024, 12, 11, 10, 0, 0, 20000),
    #     datetime(2024, 12, 11, 10, 0, 0, 30000),
    #     datetime(2024, 12, 11, 10, 0, 0, 40000),
    # ]

    result = calculate_average_milliseconds(datetime_objects)
    print(f"平均耗时（毫秒）: {result}")

    # ########### 创建一个Timeout实例，设置超时时间为5秒
    # timeout_obj = Timeout(5)
    # # 在一定时间内调用check函数进行检查
    # time.sleep(3)
    # print(timeout_obj.not_timeout())  # 预期输出True，因为还未超过5秒
    # time.sleep(3)
    # print(timeout_obj.not_timeout())  # 预期输出False，因为已经超过5秒了



import sys
import threading

import cachetools


def func_cache_ignore_args(ttl=sys.maxsize):
    cache = cachetools.TTLCache(maxsize=99999, ttl=ttl)
    lock = threading.Lock()

    def actual_decorator(func):
        def wrapper(*args, **kwargs):
            key = func.__name__
            if key in cache:
                try:
                    return cache[key]
                except KeyError:
                    pass
            with lock:
                try:
                    return cache[key]
                except KeyError:
                    pass
                result = func(*args, **kwargs)
                cache[key] = result
            return result

        return wrapper

    return actual_decorator


# 定义支持自定义ttl的缓存装饰器
def cache_with_custom_time(ttl=sys.maxsize):
    cache = cachetools.TTLCache(maxsize=99999, ttl=ttl)
    lock = threading.Lock()

    def actual_decorator(func):
        def wrapper(*args, **kwargs):
            key = args + tuple(kwargs.items())
            if key in cache:
                try:
                    return cache[key]
                except KeyError:
                    pass
            with lock:
                try:
                    return cache[key]
                except KeyError:
                    pass
                result = func(*args, **kwargs)
                cache[key] = result
            return result

        return wrapper

    return actual_decorator



import time

import cv2
import numpy as np

from auto_easy.base.ai.model_mgr_v2 import AIModelBase
from auto_easy.base.image.cvt import img_2_ndarray_rgb, img_2_list, list_2_ndarray
from auto_easy.utils import must_get_file


class SuperRes(AIModelBase):
    def __init__(self, *args, **kwargs):
        self.arg = args
        self.kwargs = kwargs

        self.scale_model_map = {}
        super().__init__(name='SuperRes',preload=False)

    def init_model(self):
        self.scale_conf = {
            3: must_get_file('EDSR_x3.pb',download_url='https://github.com/Jefung/auto_easy/raw/refs/heads/main/statics/EDSR_x3.pb'),
        }
        for scale_factor, model_path in self.scale_conf.items():
            sr = cv2.dnn_superres.DnnSuperResImpl_create()
            # 读取预训练的模型
            sr.readModel(model_path)
            # 设置放大倍数
            sr.setModel("edsr", scale_factor)
            self.scale_model_map[scale_factor] = sr

    def predict(self, img, scale_factor=3):
        if self.use_rpc():
            return self.rpc_call(img, scale_factor)

        self.wait_model_init()
        if scale_factor not in self.scale_model_map.keys():
            raise Exception('scale factor {} not supported. only support 2/3/4'.format(scale_factor))
        sr = self.scale_model_map[scale_factor]
        img = img_2_ndarray_rgb(img)
        # 对图像进行超分辨率放大
        result = sr.upsample(img)
        return result

    def rpc_req_encode(self, *args, **kwargs):
        if len(args) < 1:
            raise Exception('args length error')
        args_list = list(args)
        args_list[0] = img_2_list(args_list[0])
        args = tuple(args_list)
        return args, kwargs

    def rpc_req_decode(self, *args, **kwargs):
        args_list = list(args)
        args_list[0] = list_2_ndarray(args_list[0])
        args = tuple(args_list)
        return args, kwargs

    def rpc_resp_encode(self, origin_resp):
        return origin_resp.tolist()

    def rpc_resp_decode(self, str_resp):
        img = list_2_ndarray(str_resp)
        if img.dtype != np.uint8:
            img = img.astype(np.uint8)
        return img


# ocr = AIOCR()
if __name__ == '__main__':
    start = time.time()
    supper_res = SuperRes()
    print(time.time()-start)
    # ans = supper_res.mock_rpc_call(get_test_pic('skill_name.bmp'), scale_factor=4)
    # print(type(ans))
    # ans = ans.astype(np.uint8)
    # cv2.imshow('result', ans)
    # cv2.waitKey(0)
    # cv2.destroyAllWindows()
    # cv2.imwrite(get_test_pic('skill_name_scaled.bmp'), ans)
    # print(ans.shape)



def a_rm_ab_intersect(a, b):
    """
    从数组A中剔除A和B的交集，返回处理后的数组
    """
    return [elem for elem in a if elem not in b]


def ab_intersect(a, b):
    return [x for x in a if x in b]


def ab_union(a, b):
    result = a.copy()
    for element in b:
        if element not in result:
            result.append(element)
    return result



from auto_easy.constant import *
from auto_easy.global_log import *
from auto_easy.base import *
from auto_easy.core import *
from auto_easy.dag import *
from auto_easy.models import *
from auto_easy.utils import *
